/*
 * This is a small program to demonstrate the power of concurrency when reading
 * from disk. It scans a file for lines matching word and prints the line with
 * the word highlighted.
 *
 * Download the program like this:
 * $ curl -O https://vladov3000.com/search.c
 *
 * Compile the program like this:
 * $ cc -O2 -o search search.c
 *
 * And use it like so:
 * $ search log.txt INFO
 *
 * You can change BUFFER_COUNT to change the maximum in-flight read requests at
 * once. Note that the system limit is relatively low, with my system only
 * supporting up to 16 concurrent requests.
 *
 * You can also change the BUFFER_SIZE, which controls how much data a single
 * request fetches.
 */

#include <aio.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>

/* Terminal escape codes. Print these to change the color of the terminal. */
#define PURPLE "\x1b[35;1m"
#define RESET  "\x1b[0m"

#define BUFFER_COUNT 16
#define BUFFER_SIZE  (2 * 1024 * 1024)

#define length(array) (sizeof(array) / sizeof((array)[0]))

/* Each request has a corresponding buffer that it "owns". */
static struct aiocb requests[BUFFER_COUNT];
static char         buffers [BUFFER_COUNT][BUFFER_SIZE];

static size_t min(size_t a, size_t b) {
  return a < b ? a : b;
}

/*
 * Finds the index of the first target character in the input string after input_start.
 * Returns the input size if it is not found.
 */
static size_t find(char* input, size_t input_start, size_t input_size, char target) {
  char* pointer = memchr(&input[input_start], target, input_size - input_start);
  return pointer == NULL ? input_size : (pointer - input);
}

typedef struct {
  /* The word we are looking for. */
  char*  target;
  size_t target_size;

  /* The current line we are looking at. */
  char*  line;
  size_t line_size;
  size_t line_capacity;
} Filter;

static void print_words(Filter* filter) {
  bool   print_line  = false;
  size_t start_index = 0;

  while (start_index < filter->line_size) {
    size_t next_space = find(filter->line, start_index, filter->line_size, ' ');
    size_t next_qoute = find(filter->line, start_index, filter->line_size, '"');
    size_t end_index  = min(next_space, next_qoute);

    char*  word       = &filter->line[start_index];
    size_t word_size  = end_index - start_index;
    
    bool word_matches = false;
    if (filter->target_size == word_size) {
      word_matches = memcmp(word, filter->target, word_size) == 0;
    }

    if (word_matches) {
      if (!print_line) {
	fwrite(filter->line, 1, start_index, stdout);
      }
      print_line = true;

      fputs(PURPLE, stdout);
    }

    if (print_line) {
      fwrite(word, 1, word_size, stdout);
    }

    if (word_matches) {
      fputs(RESET, stdout);
    }

    /* Print seperator. */
    if (print_line && end_index < filter->line_size) {
      putchar(filter->line[end_index]);
    }

    start_index = end_index + 1;
  }

  if (print_line) {
    putchar('\n');
  }
}

static void feed_buffer(Filter* filter, char* buffer, size_t buffer_size) {
  size_t start_index = 0;

  while (start_index < buffer_size) {
    size_t end_index = find(buffer, start_index, buffer_size, '\n');
    size_t space     = filter->line_capacity - filter->line_size;
    size_t to_copy   = end_index - start_index;
  
    if (space < to_copy) {
      filter->line_capacity += to_copy - space;
      if (filter->line == NULL) {
	filter->line = malloc(filter->line_capacity);
      } else {
	filter->line = realloc(filter->line, filter->line_capacity);
      }
    }

    if (to_copy > 0) {
      memcpy(&filter->line[filter->line_size], &buffer[start_index], to_copy);
      filter->line_size += to_copy;
    }

    if (end_index < buffer_size) {
      print_words(filter);
      filter->line_size = 0;
    }

    start_index = end_index + 1;
  }
}

int main(int argc, char** argv) {
  if (argc != 3) {
    puts("Error: Expected exactly 2 arguments.");
    puts("Usage: search PATH TARGET");
    exit(EXIT_FAILURE);
  }

  char* path = argv[1];
  int   fd   = open(argv[1], O_RDONLY);
  assert(fd != -1);

  char* target = argv[2];

  Filter filter      = {};
  filter.target      = target;
  filter.target_size = strlen(target);
  
  struct stat info  = {};  
  int         error = fstat(fd, &info);
  assert(error == 0);

  assert(S_ISREG(info.st_mode));

  size_t file_size   = info.st_size;
  size_t file_offset = 0;

  while (file_offset < file_size) {
    size_t enqueued = 0;
    
    for (size_t i = 0; i < length(buffers); i++) {
      if (file_offset >= file_size) {
	break;
      }
      
      size_t buffer_size = BUFFER_SIZE;
      size_t remaining   = file_size - file_offset;
      if (remaining < buffer_size) {
	buffer_size = remaining;
      }
    
      struct aiocb* request = &requests[i];
      request->aio_nbytes   = BUFFER_SIZE;
      request->aio_fildes   = fd;
      request->aio_offset   = file_offset;
      request->aio_buf      = buffers[i];

      error = aio_read(request);
      if (error == -1 && errno == EAGAIN) {
	break;
      }
      assert(error == 0);
      
      enqueued    += 1;
      file_offset += buffer_size;
    }

    for (size_t i = 0; i < enqueued; i++) {
      struct aiocb* request = &requests[i];
      
      const struct aiocb* wait_for = request;
      error = aio_suspend(&wait_for, 1, NULL);
      assert(error == 0);

      ssize_t bytes_read = aio_return(request);
      assert(bytes_read != -1);

      feed_buffer(&filter, buffers[i], bytes_read);
    }
  }
}
